/*
 * CPT - a Constraint Programming Temporal planner
 *
 * File : structs.h
 *
 * Copyright (C) 2005, 2006  Vincent Vidal <vidal@cril.univ-artois.fr>
 */


#ifndef STRUCT_H
#define STRUCT_H


#include <gmp.h>
#include "var.h"
#include "pddl.h"


typedef enum {Better, Equal, Worse} Comp;


#ifndef PARSER_H
typedef struct Fluent Fluent;
typedef struct Action Action;
typedef struct Resource Resource;
#endif
typedef struct Causal Causal;
typedef union Property Property;
typedef struct ResourceData ResourceData;
typedef struct ResourceLocal ResourceLocal;
#ifdef SCHED
typedef struct ActivityConstraint ActivityConstraint;
#endif

union Property {
  char all;
  struct {
    bool consumes :1;
    bool produces :1;
    bool deletes :1;
    bool edeletes : 1;
    bool mutex :1;
    bool precedes :1;
    bool deletes_included:1;
  } bits;
};

struct Fluent {
  Num num;
  PDDLAtom *atom;
  Value *indac;
  Time *pair_cost;
  bool *mutex;
  bool is_init;
  Resource *resource;
  bool no_branching;
  //#ifdef HYBRID_PLANNER
  Time init; // pour hybride backward
  //#endif
  NORMAL_TAB(Action *, consumers);
  NORMAL_TAB(Action *, producers);
  NORMAL_TAB(Action *, deleters);
  NORMAL_TAB(Action *, edeleters);
  NORMAL_TAB(Causal *, active_causals);
  NORMAL_TAB(Causal *, causals);
};

struct Action {
  /* clonable fields */
  PDDLOperator *ope;
  NORMAL_TAB(PDDLTerm *, parameters);
  Action *origin;
  mpq_t dur_rat;
  Time dur;
  Time init;
  bool used;
  bool excluded;
  NORMAL_TAB(Fluent *, prec);
  NORMAL_TAB(Fluent *, add);
  NORMAL_TAB(Fluent *, del);
  NORMAL_TAB(Fluent *, edel);
  Num causals_nb;
  Time *distances;
  Num weight;
#ifdef SCHED
  NORMAL_TAB(ActivityConstraint *, ac_constraints);
#endif
  NORMAL_TAB(ResourceLocal *, resources);
  bool synchro;

  /* static fields */
  Num num;
  Property *properties;
  Variable start;
  Causal **causals;
  NORMAL_TAB(Action *, before);
  NORMAL_TAB(Action *, after);
  Num nb_instances;
  Variable reslevel;
};

#define CSIZE_ACTION offsetof(Action, num)

struct Causal {
  /* clonable fields */
  Fluent *fluent;
  Causal *origin;
  Num weight;

  /* static fields */
  bool required;
  bool excluded;
  Action *consumer;
  Action *best_producer;
  Variable support;
  Variable start;
};

#define CSIZE_CAUSAL offsetof(Causal, required)

#ifdef SCHED
struct ActivityConstraint {
  Fluent *fluent;
  Time min;
  mpq_t min_rat;
  Time max;
  mpq_t max_rat;
  NORMAL_TAB(Time, time);
  NORMAL_TAB(Time, dur);
};
#endif

struct Resource {
  Num num;
  PDDLAtom *atom;
  Fluent *fluent_available;
  Fluent *fluent_modified;
  Fluent *fluent_synchro;
  NORMAL_TAB(ResourceLocal *, reslocals);
};

struct ResourceLocal {
  Resource *resource;
  Action *action;
  Num index_causal;
  Time min_level;
  Time max_level;
  Time increased;
  Time decreased;
  Time assigned;
};


#define FORPROD(a, c)							\
  do { Value *_dom = (c)->support.bucket, _i = get_min(&(c)->support), _m = get_max(&(c)->support); Action **_prods = (c)->fluent->producers; \
    while (_i <= _m) { Action *a = _prods[_i]; _i = _dom[_i+1];

#define get_bits(a, b) (a)->properties[(b)->num].bits
#define get_bits_init(a, b) (a)->properties[(b)->origin->num].bits
#define get_bitsall(a, b) (a)->properties[(b)->num].all

#define set_consumes(a, f) get_bits(a, f).consumes = true
#define consumes(a,f) get_bits(a, f).consumes

#define set_produces(a, f) get_bits(a, f).produces = true
#define unset_produces(a, f) get_bits(a, f).produces = false
#define produces(a,f) get_bits(a, f).produces

#define set_deletes(a, f) NEST( get_bits(a, f).deletes = true; get_bits(a, f).edeletes = true; )
#define deletes(a, f) get_bits(a, f).deletes

#define set_edeletes(a, f) get_bits(a, f).edeletes = true
#define set_temp_edeletes(a, f) NEST( save(get_bitsall(a, f)); get_bits(a, f).edeletes = true; )
#define unset_edeletes(a, f) get_bits(a, f).edeletes = false
#define edeletes(a, f) get_bits(a, f).edeletes

#define set_amutex(a1, a2) NEST( get_bits(a1, a2).mutex = true; get_bits(a2, a1).mutex = true; )
#define amutex(a1, a2) get_bits_init(a1, a2).mutex 

/* #define set_fmutex(f1, f2) NEST( (f1)->mutex[(f2)->num] = false; (f2)->mutex[(f1)->num] = false; ) */
/* #define unset_fmutex(f1, f2) NEST( (f1)->mutex[(f2)->num] = true; (f2)->mutex[(f1)->num] = true; ) */
#define set_fmutex(f1, f2) NEST( (f1)->mutex[(f2)->num] = true; (f2)->mutex[(f1)->num] = true; )
#define unset_fmutex(f1, f2) NEST( (f1)->mutex[(f2)->num] = false; (f2)->mutex[(f1)->num] = false; )
#define fmutex(f1, f2) (f1)->mutex[(f2)->num]

#define set_precedes(a1, a2) NEST( save(get_bitsall(a1, a2)); get_bits(a1, a2).precedes = true; )
#define precedes(a1, a2) get_bits(a1, a2).precedes

#define set_distance(a1, a2, d) (a1)->distances[(a2)->num] = d
#define distance(a1, a2) (a1)->distances[(a2)->origin->num]
#define store_distance(a1, a2, d) store((a1)->distances[(a2)->num], d)

#define distance_ca(c, a) distance((c)->consumer, a)
#define distance_ac(a, c) ({Time _d = MAXCOST; FORPROD(_a2, c) { minimize(_d, distance(a, _a2)); } EFOR; _d; })

#define set_pair_cost(a1, a2, d) NEST( (a1)->pair_cost[(a2)->num] = d; (a2)->pair_cost[(a1)->num] = d; )
#define pair_cost(a1, a2) (a1)->pair_cost[(a2)->num]

#define set_deletes_included(a1, a2) NEST( get_bits(a1, a2).deletes_included = true; )
#define deletes_included(a1, a2) get_bits(a1, a2).deletes_included

#define duration(a) (a)->dur
#define first_start(a) get_inf(&(a)->start)
#define last_start(a) get_sup(&(a)->start)
#define first_end(a) (first_start(a) + duration(a))
#define last_end(a) (last_start(a) + duration(a))
#define delta_aa(a1, a2) (duration(a1) + distance(a1,a2))
#define delta_ca(c, a) delta_aa((c)->consumer, a)
#define delta_ac(a, c) (duration(a) + distance_ac(a, c))
#define slack_aa(a1, a2) (last_start(a2) - (first_start(a1) + delta_aa(a1, a2)))
#define slack_ca(c, a) slack_aa((c)->consumer, a)
#define slack_ac(a, c) (last_start(c) - (first_start(a) + delta_ac(a, c)))

#define cannot_precede_aa(a1, a2) (precedes(a2, a1) || first_start(a1) + delta_aa(a1, a2) > last_start(a2))
#define cannot_precede_ca(c, a) cannot_precede_aa((c)->consumer, a)


#define cannot_precede_ac(a, c)						\
  (cannot_precede_aa(a, (c)->consumer) ||				\
   (!can_produce(c, a) &&						\
    ({ bool _r = true;							\
      FORPROD(a2, c) {						\
	if (!precedes(a2, a) && first_start(a) + delta_aa(a, a2) <= min(last_start(a2), last_start(c))) \
	  { _r = false; break; }} EFOR; _r;})))

/* #define cannot_precede_ac(a, c) 						\ */
/*   (cannot_precede_aa(a, (c)->consumer) || (!can_produce(c, a) && first_start(a) + delta_ac(a, c) > last_start(c))) */

#define can_precede_aa(a1, a2) (first_start(a1) + delta_aa(a1, a2) <= first_start(a2))
#define can_precede_ca(c, a) can_precede_aa((c)->consumer, a)
#define can_precede_ac(a, c) (can_produce(c, a) || first_start(a) + delta_ac(a, c) <= first_start(c))

#define update_sup_a(v, x) NEST( if (x < last_start(v)) { trace_proc(update_sup_a, v, x); update_sup_variable(&(v)->start, x); } )
#define update_sup_c(v, x) NEST( if (x < last_start(v)) { trace_proc(update_sup_c, v, x); update_sup_variable(&(v)->start, x); } )
#define update_inf_a(v, x) NEST( if (x > first_start(v)) { trace_proc(update_inf_a, v, x); update_inf_variable(&(v)->start, x); } )
#define update_inf_c(v, x) NEST( if (x > first_start(v)) { trace_proc(update_inf_c, v, x); update_inf_variable(&(v)->start, x); } )

#define producer2value(c, a) (c)->fluent->indac[(a)->num]
#define value2producer(c, i) (c)->fluent->producers[i]

#define set_producer(c, a) instantiate_variable(&(c)->support, producer2value(c, a))
#define rem_producer(c, a) NEST( if (can_produce(c, a)) { trace_proc(rem_producer, c, a); remove_val_variable(&(c)->support, producer2value(c, a)); } )
#define can_produce(c, a) (!(c)->excluded && val_known(producer2value(c, a)) && can_be_instantiated_to(&(c)->support, producer2value(c, a)))
#define get_producer(c) (((c)->required && is_instantiated(&(c)->support)) ? value2producer(c, get_value(&(c)->support)) : NULL)
#define get_nbprods(c) get_card(&(c)->support)
#define activate_action(a) post_update_inf(&(a)->start);
#define activate_causal(c) NEST( if (is_instantiated(&(c)->support)) post_instantiate(&(c)->support); else post_remove_val(&(c)->support); )


#define resource_local(a, r) r->reslocals[a->origin->num]
#define min_level(a) get_inf(&(a)->reslevel)
#define max_level(a) get_sup(&(a)->reslevel)

#endif /* STRUCTS_H */
